import torch
import torch.nn as nn
from ..data.SpecialTokens import SpecialTokens

START_TOKEN, END_TOKEN, PAD_TOKEN = SpecialTokens.START_TOKEN, SpecialTokens.END_TOKEN, SpecialTokens.PAD_TOKEN


class CrfClassifier(nn.Module):
    """
    A class that defines a Conditional Random Field (CRF) classifier layer.
    The model is capable of transforming the encoder's output into the tag space, considering different strategies
    like "bilstm", "transformer", or "linear". It also includes methods for the forward algorithm, Viterbi decoding,
    and negative log likelihood calculation for training and prediction.

    Attributes:
    ----------
    label_to_id_map : dict
        Mapping of labels to their respective unique identifiers.
    input_dim : int
        Dimension of the output of the encoder.
    to_tag_space : str, optional (default: 'linear')
        The strategy to map the output of the encoder into the tag space.
        Choices: 'linear', 'bilstm', 'transformer'.
    start_label_id : int
        Identifier for the start token label.
    stop_label_id : int
        Identifier for the end token label.
    num_labels : int
        Total number of labels.
    hidden2tag : torch.nn.Linear
        Linear layer to transform the encoder's output into the tag space.
    BiLSTM : torch.nn.LSTM (optional)
        Bi-directional LSTM layer, if 'bilstm' is chosen for 'to_tag_space'.
    Transformer : torch.nn.TransformerEncoderLayer (optional)
        Transformer layer, if 'transformer' is chosen for 'to_tag_space'.
    transitions : torch.nn.Parameter
        Matrix of transition parameters, representing scores for transitioning from one state to another.

    Methods:
    ----------
    init_transition(self):
        Initializes transition constraints according to specific rules.
    _forward_alg(self, feats):
        Forward algorithm to calculate log probabilities.
    _score_sentence(self, feats, label_ids):
        Gives the score of a provided label sequence.
    _viterbi_decode(self, feats):
        Performs Viterbi decoding to find the best path, given the features.
    neg_log_likelihood(self, encoder_outputs, label_ids):
        Computes the negative log likelihood for the given encoder outputs and label sequence.
    forward(self, encoder_outputs):
        Performs forward computation to find the score and label sequence IDs, used for prediction.
    log_sum_exp_batch(log_Tensor, axis=-1):
        A static method that computes the log sum exponential operation.
    """
    def __init__(self,
                 label_to_id_map: dict,
                 input_dim: int,
                 to_tag_space: str = "linear") -> None:
        super(CrfClassifier, self).__init__()
        self.input_dim = input_dim  # dimension of output of the encoder

        # Tag space
        self.label_to_id_map = label_to_id_map
        self.start_label_id = label_to_id_map[START_TOKEN]
        self.stop_label_id = label_to_id_map[END_TOKEN]
        self.num_labels = len(label_to_id_map)

        # Maps the output of the encoder into tag space.
        self.to_tag_space = to_tag_space
        self.hidden2tag = nn.Linear(input_dim, self.num_labels)

        if to_tag_space == "bilstm":
            self.BiLSTM = nn.LSTM(bidirectional=True,
                                  num_layers=2,
                                  input_size=input_dim,
                                  hidden_size=input_dim // 2,
                                  batch_first=True)
        elif to_tag_space == "transformer":
            self.Transformer = torch.nn.TransformerEncoderLayer(d_model=input_dim,
                                                                nhead=16,
                                                                batch_first=True)
        elif to_tag_space == "bigru":
            self.BiLSTM = nn.GRU(bidirectional=True,
                                 num_layers=2,
                                 input_size=input_dim,
                                 hidden_size=input_dim // 2,
                                 batch_first=True)

        # Matrix of transition parameters.  Entry i,j is the score of
        # transitioning *to* i *from* j.
        self.transitions = nn.Parameter(
            torch.randn(self.num_labels, self.num_labels))

        self.init_transition()

    def init_transition(self) -> None:
        """
        Initializes the transition constraints for the CRF model according to specific rules regarding start, stop,
        padding tokens, and transitions between different entity tags. Enforced rules are:
        1. We never transfer to the start tag.
        2. We never transfer from the stop tag unless to pad tag.
        3. After B-x going to I-y is impossible.
        4. After O going to I-x is impossible.
        5. Going to <PAD> tag from different tag than stop tag or <PAD>.
        """
        # create entity name list
        entity_list = [entity.split("-")[-1] for entity in self.label_to_id_map.keys() if entity.startswith("B-")]

        # We never transfer to the start tag
        self.transitions.data[self.start_label_id, :] = -10000.
        # We never transfer from the stop tag unless to pad tag
        for entity in self.label_to_id_map.keys():
            if entity != PAD_TOKEN:
                self.transitions.data[self.label_to_id_map[entity], self.stop_label_id] = -10000.

        # after B-x going to I-y is impossible
        for previous_tag in entity_list:
            for next_tag in entity_list:
                if "I-"+next_tag not in self.label_to_id_map.keys() or "B-"+previous_tag not in self.label_to_id_map.keys():
                    continue
                if previous_tag != next_tag:
                    self.transitions.data[self.label_to_id_map["I-" + next_tag], self.label_to_id_map["B-" + previous_tag]] = -10000.

        # after O going to I-x is impossible
        for next_tag in entity_list:
            if "I-"+next_tag not in self.label_to_id_map.keys():
                continue
            self.transitions.data[self.label_to_id_map["I-" + next_tag], self.label_to_id_map["O"]] = -10000.

        # going to <PAD> tag from different tag than stop tag or <PAD> tag is impossible
        for previous_tag in self.label_to_id_map.keys():
            if previous_tag != PAD_TOKEN and previous_tag != END_TOKEN:
                self.transitions.data[self.label_to_id_map[PAD_TOKEN], self.label_to_id_map[previous_tag]] = -10000.

    def _forward_alg(self, feats) -> torch.Tensor:
        """
         Implements the forward algorithm to calculate log probabilities for a sequence.

        Parameters:
        ----------
        feats : torch.Tensor
            Features from the previous layer.

        Returns:
        -------
        torch.Tensor
            Log probabilities for the sequence.
        """
        # T = self.max_seq_length
        T = feats.shape[1]
        batch_size = feats.shape[0]

        # alpha_recursion,forward, alpha(zt)=p(zt,bar_x_1:t)
        log_alpha = torch.Tensor(batch_size, 1, self.num_labels).fill_(-10000.)
        # normal_alpha_0 : alpha[0]=Ot[0]*self.PIs
        # self.start_label has all of the score. it is log,0 is p=1
        log_alpha[:, 0, self.start_label_id] = 0

        # feats: sentences -> word embedding -> lstm -> MLP -> feats
        # feats is the probability of emission, feat.shape=(1,tag_size)
        for t in range(1, T):
            log_alpha = (self.log_sum_exp_batch(self.transitions + log_alpha.to(self.transitions.device), axis=-1) + feats[:, t]).unsqueeze(1)

        # log_prob of all barX
        log_prob_all_barX = self.log_sum_exp_batch(log_alpha)
        return log_prob_all_barX

    def _score_sentence(self, feats, label_ids) -> torch.Tensor:
        """
        Gives the score of a provided label sequence
        p(X=w1:t,Zt=tag1:t)=...p(Zt=tag_t|Zt-1=tag_t-1)p(xt|Zt=tag_t)...
            Gives the score of a provided label sequence based on the CRF's parameters.
        Parameters:
        ----------
        feats : torch.Tensor
            Features from the previous layer.
        label_ids : torch.Tensor
            Tensor of label IDs for the sequence.

        Returns:
        -------
        torch.Tensor
            Score for the provided label sequence.
        """

        T = feats.shape[1]
        batch_size = feats.shape[0]

        batch_transitions = self.transitions.expand(batch_size, self.num_labels, self.num_labels)
        batch_transitions = batch_transitions.flatten(1)

        score = torch.zeros((feats.shape[0], 1)).to(label_ids.device)
        # the 0th node is start_label->start_word,the probability of them=1. so t begin with 1.
        # label_ids = torch.where(label_ids == -100, torch.tensor(self.tag_to_ix[PAD_TOKEN]), label_ids)

        for t in range(1, T):
            score = score + \
                    batch_transitions.gather(-1, (label_ids[:, t] * self.num_labels + label_ids[:, t - 1]).view(-1, 1)) + \
                    feats[:, t].gather(-1, label_ids[:, t].view(-1, 1)).view(-1, 1)
        return score

    def _viterbi_decode(self, feats):
        """
        Viterbi decoding algorithm to find the best path through the CRF's states.
        Parameters:
        ----------
        feats : torch.Tensor
            Features from the previous layer.

        Returns:
        -------
        torch.Tensor
            Maximum log likelihood for the sequence.
        torch.Tensor
            The best path through the CRF's states.
        """

        T = feats.shape[1]
        batch_size = feats.shape[0]

        log_delta = torch.Tensor(batch_size, 1, self.num_labels).fill_(-10000.).to(feats.device)
        log_delta[:, 0, self.start_label_id] = 0

        # psi is for the vaule of the last latent that make P(this_latent) maximum.
        psi = torch.zeros((batch_size, T, self.num_labels), dtype=torch.long).to(feats.device) # psi[0]=0000 useless
        for t in range(1, T):
            # delta[t][k]=max_z1:t-1( p(x1,x2,...,xt,z1,z2,...,zt-1,zt=k|theta) )
            # delta[t] is the max prob of the path from  z_t-1 to z_t[k]
            log_delta, psi[:, t] = torch.max(self.transitions + log_delta, -1)
            # psi[t][k]=argmax_z1:t-1( p(x1,x2,...,xt,z1,z2,...,zt-1,zt=k|theta) )
            # psi[t][k] is the path chosen from z_t-1 to z_t[k],the value is the z_state(is k) index of z_t-1
            log_delta = (log_delta + feats[:, t]).unsqueeze(1)

        # trace back
        path = torch.zeros((batch_size, T), dtype=torch.long).to(feats.device)

        # max p(z1:t,all_x|theta)
        max_logLL_allz_allx, path[:, -1] = torch.max(log_delta.squeeze(), -1)

        for t in range(T - 2, -1, -1):
            # choose the state of z_t according the state choosed of z_t+1.
            path[:, t] = psi[:, t + 1].gather(-1, path[:, t + 1].view(-1, 1)).squeeze()

        return max_logLL_allz_allx, path

    def neg_log_likelihood(self, encoder_outputs, label_ids):
        """
        Computes the negative log likelihood for the given encoder outputs and label sequence.

        Parameters:
        ----------
        encoder_outputs : torch.Tensor
            Outputs from the encoder.
        label_ids : torch.Tensor
            Tensor of label IDs for the sequence.

        Returns:
        -------
        torch.Tensor
            Negative log likelihood.
        """
        if self.to_tag_space == "transformer":
            encoder_outputs = self.Transformer(encoder_outputs)
        elif self.to_tag_space == "bilstm":
            encoder_outputs, _ = self.BiLSTM(encoder_outputs)

        # transform to tag space
        outputs = self.hidden2tag(encoder_outputs)

        forward_score = self._forward_alg(outputs)
        # p(X=w1:t,Zt=tag1:t)=...p(Zt=tag_t|Zt-1=tag_t-1)p(xt|Zt=tag_t)...
        gold_score = self._score_sentence(outputs, label_ids)
        # - log[ p(X=w1:t,Zt=tag1:t)/p(X=w1:t) ] = - log[ p(Zt=tag1:t|X=w1:t) ]
        return torch.mean(forward_score - gold_score)

    def forward(self, encoder_outputs):
        """
        Forward computation to find the score and label sequence IDs, used for prediction.

        Parameters:
        ----------
        encoder_outputs : torch.Tensor
            Outputs from the encoder.

        Returns:
        -------
        torch.Tensor
            Score for the sequence.
        torch.Tensor
            Label sequence IDs for the sequence.
        """
        if self.to_tag_space == "transformer":
            encoder_outputs = self.Transformer(encoder_outputs)
        elif self.to_tag_space == "bilstm":
            encoder_outputs, _ = self.BiLSTM(encoder_outputs)

        # transform to tag space
        encoder_outputs = self.hidden2tag(encoder_outputs)

        # Find the best path, given the features.
        score, label_seq_ids = self._viterbi_decode(encoder_outputs)
        return score, label_seq_ids

    @staticmethod
    def log_sum_exp_batch(log_Tensor, axis=-1) -> torch.Tensor:  # shape (batch_size,n,m)
        """
        A static method that computes the log sum exponential operation for a given tensor along a specific axis.

        Parameters:
        ----------
        log_Tensor : torch.Tensor
            Input tensor.
        axis : int, optional (default: -1)
            Axis along which to compute the log sum exponential.

        Returns:
        -------
        torch.Tensor
            Result of the log sum exponential operation.
        """
        return torch.max(log_Tensor, axis)[0] + torch.log(
            torch.exp(log_Tensor - torch.max(log_Tensor, axis)[0].view(log_Tensor.shape[0], -1, 1)).sum(axis))

