import torch
from transformers import BertTokenizer, AutoTokenizer
from typing import Dict, List, Optional
from .SpecialTokens import SpecialTokens
from ..utils.download import download_file
from lambo.segmenter.lambo import Lambo
from pathlib import Path
import json

CHAR_PAD_TOKEN, CHAR_START_TOKEN, CHAR_END_TOKEN, CHAR_UNKNOWN_TOKEN, PAD_TOKEN, START_TOKEN, END_TOKEN = \
    SpecialTokens.CHAR_PAD_TOKEN, SpecialTokens.CHAR_START_TOKEN, SpecialTokens.CHAR_END_TOKEN, \
    SpecialTokens.CHAR_UNKNOWN_TOKEN, SpecialTokens.PAD_TOKEN, SpecialTokens.START_TOKEN, SpecialTokens.END_TOKEN


class NerTokenizer:
    """
       Named Entity Recognition (NER) Tokenizer.

       A tokenizer tailored for tokenizing entities and their corresponding labels.
       This tokenizer provides the flexibility to encode characters, entities, and
       labels. It can utilize pre-trained tokenizers, such as BERT and AutoTokenizer,
       and augment them with custom tokenization techniques. Based on specification passed
       in initialization it will return

       Attributes:
       ----------
       backbone_tokenizer : transformers.PreTrainedTokenizer
          The tokenizer from transformers BertTokenizer or AutoTokenizer.
       char_to_id : Dict[str, int]
           Mapping of characters to their respective unique identifiers.
       label_to_id : Dict[str, int]
           Mapping of labels to their respective unique identifiers.
       max_word_len : int
           Maximum length of word, used for padding.
       use_start_end_token : bool
           Indicates whether to append special start and end tokens to a sentence.
       tokenize_entities : bool
           Indicates whether to additionaly tokenize entities. If True, entities like "Doing" will be tokenized to "Do"
            and "##ing" to fit BERT model's vocabulary. If False, entities will be left as they are which can result
            in entities represented by BERT <unk> token.
       use_char_level_embeddings : bool
           Indicates whether to generate character-level embeddings.

       Methods:
       --------
       encode(input_entities, labels=None) -> Dict:
           Tokenizes and encodes input entities and their corresponding labels.
       encode_sentence(words: List[str]) -> Tuple:
           Encodes a sentence of words.
       _encode_labels(labels: List[str], max_word_index: int) -> torch.Tensor:
           Encodes labels into a tensor of label IDs.
       _encode_characters(words: List[str], max_word_index: int) -> torch.Tensor:
           Encodes characters of words into a tensor of character IDs.
       _create_prediction_mask(words: List[str], max_word_index: int) -> List[int]:
           Creates a prediction mask for the words.
    """
    def __init__(self,
                 pretrained_model_type: str,
                 pretrained_model_name: str,
                 char_to_id_map: Dict[str, int] = None,
                 label_to_id_map: Dict[str, int] = None,
                 use_char_level_embeddings: bool = False,
                 use_start_end_token: bool = False,
                 tokenize_entities: bool = False,
                 language: str = "pl",
                 max_word_len: int = 69,
                 load_lambo_tokenizer: bool = True):
        """
        Parameters:
        ----------
        pretrained_model_type : str
            Type of the pre-trained model to be used for tokenization (e.g., "BertModel", "AutoModel").
        pretrained_model_name : str
            Name of the pre-trained model or path.
        char_to_id_map : Dict[str, int], optional
            Mapping of characters to unique IDs.
        label_to_id_map : Dict[str, int], optional
            Mapping of labels to unique IDs.
        use_char_level_embeddings : bool, default=False
            If True, character-level embeddings will be used.
        use_start_end_token : bool, default=False
            If True, special start and end tokens will be appended.
        tokenize_entities : bool, default=False
            If True, entities will be tokenized.
         """

        assert label_to_id_map, "Specify label_to_id_map"
        assert char_to_id_map or not use_char_level_embeddings, \
            "Specify char_to_id_map if you want to use character embeddings"

        # define backbone tokenizer
        if pretrained_model_type == "BertModel":
            self.backbone_tokenizer = BertTokenizer.from_pretrained(pretrained_model_name)
        elif pretrained_model_type == "AutoModel":
            self.backbone_tokenizer = AutoTokenizer.from_pretrained(pretrained_model_name)
        else:
            raise ValueError("Not supported pretrained_model_type")

        # save mapping to ids
        self.char_to_id = char_to_id_map
        self.label_to_id = label_to_id_map

        # maximum word length
        self.max_word_len = max_word_len

        # information about what needs to be returned
        self.use_start_end_token = use_start_end_token
        self.tokenize_entities = tokenize_entities
        self.use_char_level_embeddings = use_char_level_embeddings

        # lambo segmenter
        if load_lambo_tokenizer:
            self.segmenter = Lambo.get(language, with_splitter=False)

    def encode(self,
               input_entities,
               labels: Optional[List[str]] = None,
               batchify: bool = False):
        """
        Tokenizes and encodes input entities and their corresponding labels if provided. If batchify is True returned
        dictionary with tensors will be batchified.
        """
        if isinstance(input_entities, list):
            # If input is a list of entities which happens during training this encoding is done
            input_ids, attention_mask, word_ids, max_word_index = self.encode_sentence(words=input_entities)

            output = dict(
                input_ids=input_ids,
                attention_mask=attention_mask)

            # if labels are provided we encode them
            if labels:
                label_ids = self._encode_labels(labels=labels,
                                                max_word_index=max_word_index)
                output["labels_ids"] = label_ids

            # when entities are not additionaly tokenized word_ids won't be provided
            if word_ids is not None:
                output["word_ids"] = word_ids

            # if character embeddings are used we encode words
            if self.use_char_level_embeddings:
                char_ids = self._encode_characters(words=input_entities, max_word_index=max_word_index)
                output["char_ids"] = char_ids

            output["prediction_mask"] = self._create_prediction_mask(words=input_entities,
                                                                     max_word_index=max_word_index)

            # when we use tokenizer for prediction we need to batchify data
            if batchify:
                output = {key: value.unsqueeze(0) for key, value in output.items()}

            return output
        elif isinstance(input_entities, str):
            # This is for custom prediction on strings. Segmentation using lambo has to be done before prediction
            document = self.segmenter.segment(input_entities)
            data = []
            entities = []
            for turn in document.turns:
                for sentence in turn.sentences:
                    sentence_tokens = [token.text for token in sentence.tokens]
                    entities.append(sentence_tokens)

                    # encoding
                    input_ids, attention_mask, word_ids, max_word_index = self.encode_sentence(words=sentence_tokens)

                    output = dict(
                        input_ids=input_ids,
                        attention_mask=attention_mask)

                    if word_ids is not None:
                        output["word_ids"] = word_ids

                    if self.use_char_level_embeddings:
                        char_ids = self._encode_characters(words=sentence_tokens, max_word_index=max_word_index)
                        output["char_ids"] = char_ids

                    output["prediction_mask"] = self._create_prediction_mask(words=sentence_tokens,
                                                                             max_word_index=max_word_index)
                    data.append(output)

            batched_data = {key: torch.stack([d[key] for d in data]) for key in data[0].keys()}
            return batched_data, entities
        else:
            # This is for custom prediction on strings
            raise ValueError("Not supported input type")

    def encode_sentence(self,
                        words: List[str]):
        """Transforms list of words/entities into a sequence of token ids, attentin mask for BERT model, mapping from
         tokens to words and maximal word index. Also pads data and adds special tokens if needed."""
        word_ids = None
        if not self.tokenize_entities:
            # if entities are not to be tokenized we encode them using tokenizer at once
            # this method will only work on BertTokenizer but not AutoTokenizer
            encoded_sentence = self.backbone_tokenizer.encode_plus(
                words,
                add_special_tokens=self.use_start_end_token,
                return_token_type_ids=False,
                truncation=True,
                max_length=self.backbone_tokenizer.model_max_length,
                padding="max_length",
                return_attention_mask=True,
                return_tensors="pt"
            )

            max_word_index = self.backbone_tokenizer.model_max_length - 2 if self.use_start_end_token else \
                self.backbone_tokenizer.model_max_length
        else:
            # since tokenizers doesn't provide valid mapping from token to entity we need to do it manually
            input_ids = []
            word_ids = []
            attention_mask = []

            # if start and end tokens are used we need to subtract 2 from max length of entities in sentence
            max_length = self.backbone_tokenizer.model_max_length - 2 if self.use_start_end_token else \
                self.backbone_tokenizer.model_max_length

            # enumerate words and encode them one by one
            for i, word in enumerate(words):
                word_tokenized_dict = self.backbone_tokenizer(word, add_special_tokens=False)
                if len(input_ids) + len(word_tokenized_dict["input_ids"]) > max_length:
                    # there is no need to add more tokens because they will be truncated anyway
                    break

                if len(word_tokenized_dict["input_ids"]) == 0:
                    # dealing with strange tokenizers that return empty list
                    input_ids += [self.backbone_tokenizer.unk_token_id]
                    word_ids += [i]
                    attention_mask += [1]
                else:
                    input_ids += word_tokenized_dict["input_ids"]
                    word_ids += [i] * len(word_tokenized_dict["input_ids"])  # mapping token -> word starting from 0
                    attention_mask += word_tokenized_dict["attention_mask"]

            if self.use_start_end_token:
                # adding start and end tokens
                input_ids = [self.backbone_tokenizer.cls_token_id] + input_ids + [self.backbone_tokenizer.sep_token_id]
                word_ids = [-1] + word_ids + [-1]  # they are not words so we use -1
                attention_mask = [1] + attention_mask + [1]  # attention should be applied to them

            # padding
            input_ids += [self.backbone_tokenizer.pad_token_id] *\
                         (self.backbone_tokenizer.model_max_length - len(input_ids))
            word_ids += [-1] * (self.backbone_tokenizer.model_max_length - len(word_ids))  # these are not words so we use -1
            attention_mask += [0] * (self.backbone_tokenizer.model_max_length - len(attention_mask))

            encoded_sentence = dict(
                input_ids=torch.tensor(input_ids),
                attention_mask=torch.tensor(attention_mask)
            )

            max_word_index = max(word_ids) + 1  # word ids start from 0

        if word_ids is not None:
            word_ids = torch.tensor(word_ids)

        return encoded_sentence["input_ids"].squeeze(0), encoded_sentence["attention_mask"].squeeze(0), word_ids, \
            max_word_index

    def _encode_labels(self,
                       labels: List[str],
                       max_word_index: int):
        """
        Encodes labels into a sequence of label ids and pads them
        """
        # Encoding labels and truncating if needed
        encoded_labels = [self.label_to_id[label] for label in labels[:max_word_index]]

        if self.use_start_end_token:
            # adding start and end token if start and end tokens are used
            encoded_labels = [self.label_to_id[START_TOKEN]] + encoded_labels + [self.label_to_id[END_TOKEN]]

        # Padding labels
        encoded_labels += [self.label_to_id[PAD_TOKEN]] * (self.backbone_tokenizer.model_max_length - len(encoded_labels))

        return torch.tensor(encoded_labels)

    def _encode_characters(self,
                           words: List[str],
                           max_word_index: int) -> torch.Tensor:
        """
        Encodes the characters of each word into a sequence of character ids.
        """
        # Splitting characters to create character ids
        characters = [[self.char_to_id.get(char, self.char_to_id[CHAR_UNKNOWN_TOKEN]) for char in word] for word in
                      words[:max_word_index]]

        # adding <b> and <e> tokens and truncating if characters exceed max length
        characters = [[self.char_to_id[CHAR_START_TOKEN]] + word[:self.max_word_len-2] +
                      [self.char_to_id[CHAR_END_TOKEN]] for word in characters]

        # Padding character sequence
        characters = [word + [self.char_to_id[CHAR_PAD_TOKEN]] *
                      (self.max_word_len - len(word)) for word in characters]

        # Padding word sequence
        # START token is used then add padding at the beginning
        if self.use_start_end_token:
            characters = [[self.char_to_id[CHAR_PAD_TOKEN]] * self.max_word_len] + characters

        characters += [[self.char_to_id[CHAR_PAD_TOKEN]] * self.max_word_len] * (
                    self.backbone_tokenizer.model_max_length - len(characters))

        return torch.tensor(characters)

    def _create_prediction_mask(self,
                                words: List[str],
                                max_word_index: int) -> torch.Tensor:
        """
        Creates a prediction mask for the words so when predicting we can ignore the padding and special tokens.
        """
        mask = [1]*min(len(words), max_word_index)

        if self.use_start_end_token:
            mask = [0] + mask + [0]

        # padding
        mask += [0] * (self.backbone_tokenizer.model_max_length - len(mask))
        return torch.tensor(mask)

    def decode(self, ids):
        """
        Decode a list of IDs or list of lists of IDs to their corresponding labels.
        """
        # if predicted tag is PAD or START or END we return O tag
        id_to_label = {v: (k if k not in [PAD_TOKEN, END_TOKEN, START_TOKEN] else "O")
                       for k, v in self.label_to_id.items()}

        if len(ids) == 0:
            return []

        # Check if it's a list of lists
        if isinstance(ids[0], list):
            return [[id_to_label[i] for i in inner_list] for inner_list in ids]
        else:
            return [id_to_label[i] for i in ids]

    @classmethod
    def get(cls,
            model_name: str,
            load_lambo_tokenizer : bool = True):
        """
        Downloads and loads pretrained tokenizer from MinIO
        """
        # Downloading files and loading jsons
        PATH_CONFIG = Path(download_file(model_name, "config.json"))
        config = json.load(open(PATH_CONFIG))

        try:
            PATH_CHAR_TO_ID = Path(download_file(model_name, "char_to_id.json"))
            char_to_id = json.load(open(PATH_CHAR_TO_ID))
        except:
            char_to_id = {}

        PATH_LABEL_TO_ID = Path(download_file(model_name, "label_to_id.json"))
        label_to_id = json.load(open(PATH_LABEL_TO_ID))

        # load tokenizer
        tokenizer = NerTokenizer(pretrained_model_type=config["model"]["bert_embedder"]["pretrained_model_type"],
                                 pretrained_model_name=config["model"]["bert_embedder"]["pretrained_model_name"],
                                 char_to_id_map=char_to_id,
                                 label_to_id_map=label_to_id,
                                 use_char_level_embeddings=config["data"]["use_char_level_embeddings"],
                                 use_start_end_token=config["data"]["use_start_end_token"],
                                 tokenize_entities=config["data"]["tokenize_entities"],
                                 language=config["data"].get("language", "pl"),
                                 max_word_len=config["data"].get("max_word_len", 69),
                                 load_lambo_tokenizer=load_lambo_tokenizer)

        return tokenizer

    @classmethod
    def load_from_disc(cls,
                       folder_path: Path,
                       load_lambo_tokenizer : bool = True):
        """
        Loads pretrained tokenizer from disc
        """
        # Loading jsons
        PATH_CONFIG = folder_path / "config.json"
        config = json.load(open(PATH_CONFIG))

        try:
            PATH_CHAR_TO_ID = folder_path / "char_to_id.json"
            char_to_id = json.load(open(PATH_CHAR_TO_ID))
        except:
            char_to_id = {}

        PATH_LABEL_TO_ID = folder_path / "label_to_id.json"
        label_to_id = json.load(open(PATH_LABEL_TO_ID))

        # load tokenizer
        tokenizer = NerTokenizer(pretrained_model_type=config["model"]["bert_embedder"]["pretrained_model_type"],
                                 pretrained_model_name=config["model"]["bert_embedder"]["pretrained_model_name"],
                                 char_to_id_map=char_to_id,
                                 label_to_id_map=label_to_id,
                                 use_char_level_embeddings=config["data"]["use_char_level_embeddings"],
                                 use_start_end_token=config["data"]["use_start_end_token"],
                                 tokenize_entities=config["data"]["tokenize_entities"],
                                 language=config["data"].get("language", "pl"),
                                 max_word_len=config["data"]["max_word_len"],
                                 load_lambo_tokenizer=load_lambo_tokenizer)
        return tokenizer
