import torch
from torch import nn


class TokenPooling(nn.Module):
    """
       TokenPooling Module.

       This module is used for pooling the token embeddings obtained from BERT models.
       It provides two pooling strategies: max and mean. It is designed to transform token
       embeddings into word embeddings by applying the selected pooling strategy over the
       tokens of each word. The class also supports the option to account for special start
       and end tokens.

       Attributes:
       -----------
       pooling_strategy : str, optional
           The strategy to pool token embeddings. Must be one of ['max', 'mean']. Default is 'max'.
       use_start_end_token : bool, optional
           Whether to consider special start and end tokens in pooling. Default is False.

       Methods:
       --------
       forward(bert_outputs: torch.Tensor, word_ids: torch.Tensor) -> torch.Tensor:
           Applies the specified pooling strategy on the token embeddings and returns the word embeddings.
    """
    def __init__(self,
                 pooling_strategy: str = 'max',
                 use_start_end_token: bool = False):
        super(TokenPooling, self).__init__()
        
        assert pooling_strategy in ['max', 'mean'], "Pooling strategy must be one of ['max', 'mean']"
        
        self.pooling_strategy = pooling_strategy
        self.use_start_end_token = use_start_end_token
    
    def forward(self, 
                bert_outputs: torch.Tensor,
                word_ids: torch.Tensor) -> torch.Tensor:
        """
        Forward method for pooling token embeddings.
        This method pools the token embeddings into word embeddings using the selected pooling strategy
        ('max' or 'mean'). If use_start_end_token is set to True, the method will consider special start
        and end tokens during pooling.

        Parameters
        ----------
        bert_outputs: torch.Tensor
            Token embeddings of shape (batch_size, max_seq_len, hidden_size).
        word_ids: torch.Tensor
            Word ID mapping of shape (batch_size, max_seq_len).

        Returns
        ----------
        torch.Tensor
            Pooled word embeddings of shape (batch_size, max_seq_len, hidden_size).
        """
        # Max pooling of token embeddings to obtain word embeddings
        # bert_outputs which are token embeddings: (batch_size, max_seq_len, hidden_size)
        # word_ids: (batch_size, max_seq_len)

        # info whether to move word embeddings idx
        # first token may be [CLS]
        if self.use_start_end_token:
            word_ids = word_ids + 1

        word_ids = word_ids.unsqueeze(2).expand_as(bert_outputs)
        word_ids = word_ids.type(torch.int64)
        pooled = torch.zeros(bert_outputs.size(0), word_ids.max() + 1, bert_outputs.size(2)).to(bert_outputs.device)

        if self.pooling_strategy == "max":
            pooled = pooled.scatter_reduce(1, word_ids, bert_outputs, reduce="amax", include_self=False)
        elif self.pooling_strategy == "mean":
            pooled = pooled.scatter_reduce(1, word_ids, bert_outputs, reduce="mean", include_self=False)

        if self.use_start_end_token:
            pooled[:, 0, :] = 0
            # maybe adding cls and sep token embeddings are needed here

            # expanding with zeros so dimensions are maintained
        rows_to_add = bert_outputs.shape[1] - pooled.shape[1]
        zeros_tensor = torch.zeros(bert_outputs.shape[0], rows_to_add, bert_outputs.shape[2]).to(bert_outputs.device)

        # word_embeddings are pooled embeddings of tokens across word indexes
        # fot other tokens like START, END, PAD they are 0s
        # word_embeddings: (batch_size, max_seq_len, hidden_size)
        bert_outputs = torch.cat((pooled, zeros_tensor), dim=1)

        return bert_outputs

