from typing import Optional, Sequence
import torch
from torch import Tensor
from torch import nn
from torch.nn import functional as F


class FocalLoss(nn.Module):
    """ Focal Loss, as described in https://arxiv.org/abs/1708.02002.

    It is essentially an enhancement to cross entropy loss and is
    useful for classification tasks when there is a large class imbalance.
    x is expected to contain raw, unnormalized scores for each class.
    y is expected to contain class labels.

    Shape:
        - x: (batch_size, C) or (batch_size, C, d1, d2, ..., dK), K > 0.
        - y: (batch_size,) or (batch_size, d1, d2, ..., dK), K > 0.

    Attributes:
    ----------
    alpha : Tensor, optional
        Weights for each class. Default is None.
    gamma : float
        A constant, as described in the referenced paper. Default is 0.
    ignore_index : int
        Class label to ignore. Default is -100.
    reduction : str
        Specifies the reduction to apply to the output: 'mean', 'sum', or 'none'.
    nll_loss : torch.nn.modules.loss.NLLLoss
        Negative log likelihood loss.

    Methods:
    -------
    forward(x: Tensor, y: Tensor) -> Tensor
        Compute the focal loss for the given inputs.

    """

    def __init__(self,
                 alpha: Optional[Tensor] = None,
                 gamma: float = 2.,
                 reduction: str = 'mean',
                 ignore_index: int = -100):
        """Constructor.

        Parameters:
        ----------
        alpha : Tensor, optional
            Weights for each class. Default is None.
        gamma : float, optional
            A constant, as described in the referenced paper. Default is 0.
        reduction : str, optional
            Specifies the reduction to apply to the output: 'mean', 'sum', or 'none'. Default is 'mean'.
        ignore_index : int, optional
            Class label to ignore. Default is -100.
        """
        if reduction not in ('mean', 'sum', 'none'):
            raise ValueError(
                'Reduction must be one of: "mean", "sum", "none".')

        super().__init__()
        self.alpha = alpha
        self.gamma = gamma
        self.ignore_index = ignore_index
        self.reduction = reduction

        self.nll_loss = nn.NLLLoss(
            weight=alpha, reduction='none', ignore_index=ignore_index)

    def __repr__(self):
        arg_keys = ['alpha', 'gamma', 'ignore_index', 'reduction']
        arg_vals = [self.__dict__[k] for k in arg_keys]
        arg_strs = [f'{k}={v!r}' for k, v in zip(arg_keys, arg_vals)]
        arg_str = ', '.join(arg_strs)
        return f'{type(self).__name__}({arg_str})'

    def forward(self, x: Tensor, y: Tensor) -> Tensor:
        if x.ndim > 2:
            # (N, C, d1, d2, ..., dK) --> (N * d1 * ... * dK, C)
            c = x.shape[1]
            x = x.permute(0, *range(2, x.ndim), 1).reshape(-1, c)
            # (N, d1, d2, ..., dK) --> (N * d1 * ... * dK,)
            y = y.view(-1)

        unignored_mask = y != self.ignore_index
        y = y[unignored_mask]
        if len(y) == 0:
            return torch.tensor(0.)
        x = x[unignored_mask]

        # compute weighted cross entropy term: -alpha * log(pt)
        # (alpha is already part of self.nll_loss)
        log_p = F.log_softmax(x, dim=-1)
        ce = self.nll_loss(log_p, y)

        # get true class column from each row
        all_rows = torch.arange(len(x))
        log_pt = log_p[all_rows, y]

        # compute focal term: (1 - pt)^gamma
        pt = log_pt.exp()
        focal_term = (1 - pt)**self.gamma

        # the full loss: -alpha * ((1 - pt)^gamma) * log(pt)
        loss = focal_term * ce

        if self.reduction == 'mean':
            loss = loss.mean()
        elif self.reduction == 'sum':
            loss = loss.sum()

        return loss


def focal_loss(alpha: Optional[Sequence] = None,
               gamma: float = 2.,
               reduction: str = 'mean',
               size: int = 10,
               ignore_index: int = -100,
               device='cpu',
               dtype=torch.float32) -> FocalLoss:
    """Factory function for FocalLoss.

    Parameters:
    ----------
    alpha : Sequence, optional
        Weights for each class. If provided, it will be converted to a Tensor. Default is None.
    gamma : float, optional
        A constant, as described in the referenced paper. Default is 0.
    reduction : str, optional
        Specifies the reduction to apply to the output: 'mean', 'sum', or 'none'. Default is 'mean'.
    size : int, optional
        Number of classes.
    ignore_index : int, optional
        Class label to ignore. Default is -100.
    device : str, optional
        Device to move the alpha parameter to. Default is 'cpu'.
    dtype : torch.dtype, optional
        Data type to cast the alpha parameter to. Default is torch.float32.

    Returns:
    -------
    FocalLoss
        A configured FocalLoss object.
    """
    if alpha is not None:
        if not isinstance(alpha, Tensor):
            alpha = torch.full((size,), alpha)
        alpha = alpha.to(device=device, dtype=dtype)

    fl = FocalLoss(
        alpha=alpha,
        gamma=gamma,
        reduction=reduction,
        ignore_index=ignore_index)
    return fl
