from typing import Dict
import warnings
import string
import random
from .default_config import Config
import torch
import warnings


def fix_common_warnings(float_precision="medium"):
    """Fix common warnings that are not relevant to the user"""
    # Fixing
    # You are using a CUDA device (<graphic-card>) that has Tensor Cores.To properly utilize them, you should set ...
    torch.set_float32_matmul_precision(float_precision)

    # Fixing warning about number of workers being too small.
    warnings.filterwarnings("ignore", ".*does not have many workers.*")

    # Fixing warning about depreciation of tensorboard
    warnings.filterwarnings("ignore", ".*`tensorboardX` has been removed as a dependency*")


def move_tensors_to_device(tensor_dict, device):
    """Move all tensors in a dict to a device"""
    return {key: tensor.to(device) for key, tensor in tensor_dict.items()}


def generate_random_string(length=16):
    """Generate a random string of given length"""
    letters = string.ascii_lowercase
    return ''.join(random.choice(letters) for i in range(length))


def custom_formatwarning(msg, *args, **kwargs):
    """Custom warning formatter"""
    # ignore everything except the message
    return "WARNING: " + str(msg) + '\n'


def check_config_constraints(config) -> Dict:
    """Check if the config is valid according to the constraints and fixes config if possible"""
    POSSIBLE_LOSSES = ["crf", "ce", "focal"]

    warnings.formatwarning = custom_formatwarning

    # loading default config
    default_config = Config.default_config

    data_config = config.get("data", {})
    model_config = config.get("model", {})
    trainer_config = config.get("trainer", {})

    # Check if learning rate is specified
    if "learning_rate" not in config.keys():
        warnings.warn(f"Learning rate is not specified")
    if "loss" not in config.keys():
        warnings.warn(f"Loss is not specified")
    else:
        if config["loss"] not in POSSIBLE_LOSSES:
            raise Exception(f"Loss is not in {POSSIBLE_LOSSES}")

    # Check if the model config is valid
    if model_config:
        # check bert embedder
        bert_embedder_config = model_config.get("bert_embedder", {})
        bert_embedder_config_keys = bert_embedder_config.keys()
        if "pretrained_model_name" not in bert_embedder_config_keys:
            raise Exception("Specify pretrained_model_name in model config")

        if "pretrained_model_type" not in bert_embedder_config_keys:
            raise Exception("Specify pretrained_model_type in model config")

        if "freeze_bert" not in bert_embedder_config_keys:
            default_value = default_config["model"]["bert_embedder"]["freeze_bert"]
            warnings.warn(f"No freeze_bert specified, using default value : {default_value}")
            bert_embedder_config["freeze_bert"] = default_value

        if "token_pooling" not in bert_embedder_config_keys:
            default_value = default_config["model"]["bert_embedder"]["token_pooling"]
            warnings.warn(f"No token_pooling specified, using default value - {default_value}")
            bert_embedder_config["token_pooling"] = default_value

        if bert_embedder_config["token_pooling"]:
            warnings.warn("Token pooling is enabled, setting data-tokenize_entities to True")
            data_config["tokenize_entities"] = True
        else:
            warnings.warn("Token pooling is disables, setting data-tokenize_entities to False")
            data_config["tokenize_entities"] = False

        if "pooling_strategy" not in bert_embedder_config_keys and bert_embedder_config["token_pooling"]:
            default_value = default_config["model"]["bert_embedder"]["pooling_strategy"]
            warnings.warn(f"No pooling_strategy specified, using default value: {default_value}")
            bert_embedder_config["pooling_strategy"] = default_value
        else:
            if bert_embedder_config["token_pooling"] and bert_embedder_config["pooling_strategy"] not in ["max", "mean"]:
                raise Exception("Specify pooling_strategy in model config - possible values [max, mean]")

        config["model"]["bert_embedder"] = bert_embedder_config

        # check char_embedder
        char_embedder_config = model_config.get("char_embedder", {})
        char_embedder_config_keys = char_embedder_config.keys()
        if char_embedder_config:
            if char_embedder_config.get("type", "-") not in ["combo", "contextual"]:
                raise Exception("Specify char_embedder type in model config - possible values [combo, contextual]")

            if "char_embedding_dim" not in char_embedder_config_keys:
                default_value = default_config["model"]["char_embedder"]["char_embedding_dim"]
                warnings.warn(f"No char_embedding_dim specified, using default value {default_value}")
                char_embedder_config["char_embedding_dim"] = default_value

            warnings.warn("Character embedder architecture specified, setting data-use_char_level_embeddings to True")
            data_config["use_char_level_embeddings"] = True
        else:
            warnings.warn("Character embedder architecture not specified, setting data-use_char_level_embeddings to False")
            data_config["use_char_level_embeddings"] = False

        config["model"]["char_embedder"] = char_embedder_config

        # check classifier
        classifier_config = model_config.get("classifier", {})
        classifier_config_keys = classifier_config.keys()
        if classifier_config:
            if classifier_config.get("type", "-") not in ["crf", "vanilla"]:
                raise Exception("Specify classifier type in model config - possible values [crf, vanilla]")
            if classifier_config.get("to_tag_space", "-") not in ["linear", "bilstm", "transformer"]:
                raise Exception("Specify transformation to tag space in model config - possible values [linear, bilstm, transformer]")
        else:
            raise Exception("Specify model classifier in model config")

    else:
        raise Exception("Specify model architecture in model config")

    # Check if the data config is valid
    data_config_keys = data_config.keys()
    if "use_char_level_embeddings" not in data_config_keys:
        default_value = default_config["data"]["use_char_level_embeddings"]
        warnings.warn(f"No use_char_level_embeddings specified, using default value {default_value}")
        data_config["use_char_level_embeddings"] = default_value

    if "use_start_end_token" not in data_config_keys:
        default_value = default_config["data"]["use_start_end_token"]
        warnings.warn(f"No use_start_end_token specified, using default value {default_value}")
        data_config["use_start_end_token"] = default_value

    if "tokenize_entities" not in data_config_keys:
        default_value = default_config["data"]["tokenize_entities"]
        warnings.warn(f"No tokenize_entities specified, using default value {default_value}")
        data_config["tokenize_entities"] = default_value

    if "batch_size" not in data_config_keys:
        default_value = default_config["data"]["batch_size"]
        warnings.warn(f"No batch_size specified, using default value {default_value}")
        data_config["batch_size"] = default_value

    if "encoding" not in data_config_keys:
        default_value = default_config["data"]["encoding"]
        warnings.warn(f"No encoding specified, using default value {default_value}")
        data_config["encoding"] = default_value

    if "num_workers" not in data_config_keys:
        default_value = default_config["data"]["num_workers"]
        warnings.warn(f"No num_workers specified, using default value {default_value}")
        data_config["num_workers"] = default_value

    config["data"] = data_config

    # Check if the training config is valid
    if trainer_config:
        pass
    else:
        warnings.warn("No trainer config provided, using default trainer config")
        config["trainer"] = default_config["trainer"]

    return config
