"""
This script facilitates the evaluation of prediction process for a pre-trained Named Entity Recognition (NER) model.
At the end of evaluation it prints classification report and global metrics.

Arguments:
---------
--model_path : str, optional (default: "./models/pl_example")
Path to the folder containing the pre-trained NER model and associated files.
--data_file_path : str, optional (default: "./data/pl/test.txt")
Path to the dataset file that needs to be fed to the model for prediction.
--device : int, optional (default: 0)
The computational device to be used. '-1' signifies CPU usage. Any other integer denotes the specific CUDA device.
--batch_size : int, optional (default: 24)
The number of data points to be processed in each batch during prediction.
--encoding : str, optional (default: "utf-8")
Encoding format used to read the dataset file.
"""
import argparse
from pathlib import Path
import json
import torch
from combo.ner_modules.NerModel import NerModel
from combo.ner_modules.data.NerDataset import NerDataset
from torch.utils.data import DataLoader
from combo.ner_modules.metrics.MetricHolder import MetricHolder
from tqdm import tqdm
from combo.ner_modules.utils.constructors import construct_loss_from_config, construct_tokenizer_from_config
from combo.ner_modules.utils.utils import fix_common_warnings

# Argument parsing
parser = argparse.ArgumentParser()
parser.add_argument("--model_path", action="store", default="./models/pl_example", help="Path folder with model")
parser.add_argument("--data_file_path", action="store", default="./data/pl/test.txt", help="Path to file with data")
parser.add_argument("--device", action="store", default=0, help="Specify device")
parser.add_argument("--batch_size", action="store", default=24, help="Specify device")
parser.add_argument("--encoding", action="store", default="utf-8", help="Specify device")


args = parser.parse_args()
PATH_MODEL = Path(args.model_path)
PATH_DATA_FILE = Path(args.data_file_path)


def move_tensors_to_device(tensor_dict, device):
    return {key: tensor.to(device) for key, tensor in tensor_dict.items()}


if __name__ == "__main__":
    fix_common_warnings()

    # load jsons
    config = json.load(open(PATH_MODEL / "config.json"))
    try:
        char_to_id = json.load(open(PATH_MODEL / "char_to_id.json"))
    except:
        char_to_id = {}
    label_to_id = json.load(open(PATH_MODEL / "label_to_id.json"))

    # specify device
    if int(args.device) == -1:
        device = torch.device("cpu")
    else:
        device = torch.device(f"cuda:{args.device}")

    # define model
    loss_fn = construct_loss_from_config(config=config,
                                         label_to_id=label_to_id)

    # load model
    model = NerModel.load_from_checkpoint(PATH_MODEL / "best_model.ckpt",
                                          loss_fn=loss_fn,
                                          char_to_id_map=char_to_id,
                                          label_to_id_map=label_to_id,
                                          config=config)
    model.to(device)

    # create dataset & dataloader
    tokenizer = construct_tokenizer_from_config(config=config,
                                                char_to_id_map=char_to_id,
                                                label_to_id_map=label_to_id)

    dataset = NerDataset(file_path=PATH_DATA_FILE,
                         tokenizer=tokenizer,
                         encoding=args.encoding)

    dataloader = DataLoader(dataset,
                            batch_size=args.batch_size,
                            shuffle=False)

    # predicting
    metrics_holder = MetricHolder(label_to_idx=label_to_id)
    model.freeze()
    i = 0
    for batch in tqdm(dataloader):
        batch = move_tensors_to_device(batch, device)
        model.test_step(batch, i)
        i += 1

    print("#" * 50)
    print("Classification report")
    print(model.test_metrics_holder.get_classification_report())
    print("General metrics")
    metrics = model.test_metrics_holder.get_metrics(stage="eval")
    for metric in metrics:
        print(f"{metric}: {metrics[metric]}")







