"""
This script allows for the training and testing of a Named Entity Recognition model using PyTorch Lightning.

Arguments:
---------
--config_path : str, optional (default: "./configs/default_config.json")
    Path to the configuration file used for training the model.
--n_reruns : int, optional (default: 1)
    Number of model reruns.
--data_path : str, optional (default: "")
    Path to the data, if not provided, taken from the config.
--serialization_dir : str, optional (default: "./models/")
    Path to save the model.
--check_config : bool, optional
    Flag to check the constraints on the configuration.
--use_wandb_logger : bool, optional
    Flag to determine whether to use the wandb logger.
--wandb_project_name : str, optional (default: "NER_ipi_pan3")
    Name of the WANDB project.
"""
import argparse
from pathlib import Path
import torch
import json
import shutil
import os
import pytorch_lightning as pl
from combo.ner_modules.NerModel import NerModel
from combo.ner_modules.data.utils import create_tag2id, create_char2id, calculate_longest_word
from pytorch_lightning.loggers import WandbLogger
import wandb
from combo.ner_modules.utils.utils import generate_random_string, check_config_constraints
from combo.ner_modules.utils.constructors import construct_callbacks_from_config, construct_loss_from_config, construct_tokenizer_from_config,\
    construct_data_module_from_config
from combo.ner_modules.utils.utils import fix_common_warnings

# Argument parsing
parser = argparse.ArgumentParser()
parser.add_argument("--config_path", action="store", default="./configs/default_config.json", help="Path to config file")
parser.add_argument("--n_reruns", action="store", default=1, help="Number of model reruns")
parser.add_argument("--data_path", action="store", default="", help="Path to data, if not provided taken from config")
parser.add_argument("--serialization_dir", action="store", default="./models/", help="Path to save_model")
parser.add_argument("--check_config", action="store_true", help="Flag whether check on config should be done")
parser.add_argument("--use_wandb_logger", action="store_false", help="Flag whether to use wandb logger")
parser.add_argument("--wandb_project_name", action="store", default="NER_ipi_pan3", help="WANDB project name")


args = parser.parse_args()

PATH_CONFIG_FILE = Path(args.config_path)
N_RERUNS = int(args.n_reruns)
DATA_PATH = args.data_path

torch.set_float32_matmul_precision("medium")  # to make lightning happy

if __name__ == "__main__":
    # fixing common warnings
    fix_common_warnings()

    # Loading config file
    default_config = open(PATH_CONFIG_FILE)
    config = json.load(default_config)

    # Wandb logger config name will be used as run name
    config["name"] = PATH_CONFIG_FILE.stem

    # Parse data path to config if it is given
    if DATA_PATH:
        config["data"]["path_data"] = DATA_PATH

    # Check config
    if args.check_config:
        config = check_config_constraints(config)

    # create vocabularies
    char_to_id = create_char2id(file_path=Path(config["data"]["path_data"]) / "train.txt",
                                encoding=config["data"]["encoding"])
    label_to_id = create_tag2id(file_path=Path(config["data"]["path_data"]) / "train.txt",
                                encoding=config["data"]["encoding"],
                                include_special_tokens=config["data"]["use_start_end_token"])

    # Extract max word length
    config["data"]["max_word_len"] = calculate_longest_word(file_path=Path(config["data"]["path_data"]) / "train.txt",
                                                            encoding=config["data"]["encoding"])

    # construct tokenizer
    tokenizer = construct_tokenizer_from_config(config=config,
                                                char_to_id_map=char_to_id,
                                                label_to_id_map=label_to_id)
    # construct data module
    data_module = construct_data_module_from_config(config=config,
                                                    tokenizer=tokenizer)

    # construct loss
    loss = construct_loss_from_config(config=config,
                                      label_to_id=label_to_id)

    for k in range(N_RERUNS):
        print("#"*50)
        print(f"Starting training number {k+1} out of {N_RERUNS}")
        serialization_dir = Path(args.serialization_dir) / generate_random_string()
        serialization_dir.mkdir(parents=True, exist_ok=True)

        if args.use_wandb_logger:
            wandb.login()
            logger = WandbLogger(project=args.wandb_project_name,
                                 config=config,
                                 reinit=True)
        else:
            logger = True

        # save vocab
        with open(serialization_dir / "char_to_id.json", "w+") as f:
            json.dump(char_to_id, f)

        with open(serialization_dir / "label_to_id.json", "w+") as f:
            json.dump(label_to_id, f)

        with open(serialization_dir / "config.json", "w+") as f:
            json.dump(config, f)

        model = NerModel(loss_fn=loss,
                         char_to_id_map=char_to_id,
                         label_to_id_map=label_to_id,
                         config=config)

        # initialize trainer
        params = config["trainer"]
        params["callbacks"] = construct_callbacks_from_config(config.get("callbacks", {}))
        params["default_root_dir"] = serialization_dir
        params["logger"] = logger
        trainer = pl.Trainer(**params)

        # start training
        trainer.fit(model,
                    datamodule=data_module)

        # Testing
        if (Path(config["data"]["path_data"]) / "test.txt").is_file():
            test_result = trainer.test(verbose=True,
                                       ckpt_path='best',
                                       datamodule=data_module)

        if args.use_wandb_logger:
            wandb.finish()

        # Moving best model to correct directory
        if "ModelCheckpoint" in config.get("callbacks", {}).keys():
            checkpoint_callback = params["callbacks"][-2]
            best_model_path = checkpoint_callback.best_model_path
            if not best_model_path:
                best_model_path = f"./{logger.name}/{logger.version}/checkpoints/{checkpoint_callback.filename}.ckpt"
            new_name = 'best_model.ckpt'
            new_path = os.path.join(serialization_dir, new_name)
            shutil.move(best_model_path, new_path)

        print(f"Serialization directory: {serialization_dir}")
